#!/usr/bin/env python

'''Super-Simple Scheduler for BG/L'''
__revision__ = '$Revision'

from copy import deepcopy
from sss.server import Server
from sss.restriction import Data, DataSet
from elementtree.ElementTree import XML, Element
from syslog import syslog, LOG_INFO, LOG_ERR
from time import time
import sys

sys.path.append('/soft/apps/rm-0.90/lib/python')
import DB2

class DataSet2(DataSet):
    '''DataSet2 is DataSet with class parameters'''
    rname = None
    ename = None
    obj = None
    idalloc = None
    log = False
    
    def __init__(self):
        self.data = []

class FailureMode(object):
    '''FailureModes are used to report (and supress) errors appropriately
    call Pass() on success and Fail() on error'''
    def __init__(self, name):
        self.name = name
        self.status = True

    def Pass(self):
        '''Check if status was previously failure and report OK status if needed'''
        if not self.status:
            syslog(LOG_ERR, "Failure %s cleared" % (self.name))
            self.status = True

    def Fail(self):
        '''Check if status was previously success and report failed status if needed'''
        if self.status:
            syslog(LOG_ERR, "Failure %s occured" % (self.name))
            self.status = False

class Partition(Data):
    '''Partitions are allocatable chunks of the machine'''
    def __init__(self, element):
        Data.__init__(self, element)
        self.element.set('state', 'idle')
        self.job = 'none'
        self.rcounter = 1

    #def __cmp__(self, other):
    #    return int(self.element.get('size')).__cmp__(int(other.element.get('size')))

    def CanRun(self, job):
        '''Check that job can run on partition with reservation constraints'''
        if job.element.get('queue') not in self.element.get('queue').split(':') + ['BUG']:
            return False
        jobsize = int(job.element.get('nodes'))
        partsize = int(self.element.get('size'))
        if jobsize > partsize:
            return False
        if (((jobsize * 2) <= partsize) and (partsize != 32)):
            # job should be run on a smaller partition
            return False
        # add a little slack for job cleanup with reservation calculations
        wall = float(job.element.get('walltime')) + 5.0
        jdur = 60 * wall
        # all times are in seconds
        current = time()
        for reserv in self.element.findall('Reservation'):
            if job.element.get('user') in reserv.get('user', '').split(':'):
                continue
            start = float(reserv.get('start'))
            rdur = int(reserv.get('duration'))
            if current < start:
                # reservation has not started
                if start < (current + jdur):
                    return False
            elif current > (start + rdur):
                # reservation has finished
                continue
            else:
                # reservation is active
                return False
        return True

    def PlaceJob(self, job):
        '''Allocate this partition for Job'''
        syslog(LOG_INFO, "Scheduling job %s on partition %s" % (job.element.get('jobid'),
                                                                self.element.get('name')))
        self.job = job.element.get('jobid')
        self.element.set('state', 'busy')

    def Free(self):
        '''DeAllocate partition for current job'''
        syslog(LOG_INFO, "Freeing partition %s" % (self.element.get('name')))
        self.job = 'none'
        self.element.set('state', 'idle')

    def AddReservation(self, args):
        '''Add a reservation for this partition'''
        reservation = Element('Reservation')
        reservation.attrib.update(args)
        if not args.has_key('name'):
            reservation.set('name', "%s.%s" % (self.element.get('name'), self.rcounter))
            self.rcounter += 1
        self.element.append(reservation)

    def DelReservation(self, name):
        [self.element.remove(reservation) for reservation in self.element.findall('Reservation') if reservation.get('name') == name]

class Job(Data):
    '''This class is represents User Jobs'''
    def __init__(self, element):
        Data.__init__(self, element)
        self.partition = 'none'
        syslog(LOG_INFO, "Found new job %s" % (self.element.get('jobid')))

    def Place(self, partition):
        '''Build linkage to execution partition'''
        self.partition = partition.element.get('name')
        self.element.set('state', 'running')
        

class PartitionSet(DataSet2):
    rname = 'Partitions'
    ename = 'Partition'
    obj = Partition

    def __init__(self):
        DataSet2.__init__(self)
        self.db2 = DB2.connect(uid='bglsysdb', pwd='db24bgls', dsn='bgdb0').cursor()
        self.jobs = []
        self.qmconnect = FailureMode("QM Connection")

    def __getstate__(self):
        return {'data':deepcopy(self.data), 'jobs':deepcopy(self.jobs)}

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.db2 = DB2.connect(uid='bglsysdb', pwd='db24bgls', dsn='bgdb0').cursor()

    def Schedule(self, jobs):
        '''Find new jobs, fit them on a partitions'''
        knownjobs = [job.element.get('jobid') for job in self.jobs]
        activejobs = [job.get('jobid') for job in jobs]
        finished = [jobid for jobid in knownjobs if jobid not in activejobs]
        #print "known", knownjobs, "active", activejobs, "finished", finished
        # add new jobs
        [self.jobs.append(Job(jobdata)) for jobdata in jobs if jobdata.get('jobid') not in knownjobs]
        # delete finished jobs
        [self.jobs.remove(job) for job in self.jobs if job.element.get('jobid') in finished]
        # free partitions with nonexistant jobs
        [partition.Free() for partition in self.data if partition.job not in activejobs + ['none']]
        # find idle partitions for new jobs
        candidates = [part for part in self.data if part.element.get('state') == 'idle' and
                      part.element.get('admin') == 'online']
        #print "candidates: ", [cand.element.get('name') for cand in candidates]
        # find idle jobs
        idlejobs = [job for job in self.jobs if job.element.get('state') == 'queued']
        #print "jobs:", self.jobs
        #print "idle jobs:", [idlej.element.get('jobid') for idlej in idlejobs]
        #print "not idle:", [(nidlej.element.get('jobid'), nidlej.element.get('state')) for nidlej in self.jobs if nidlej not in idlejobs]
        if candidates and idlejobs:
            self.db2.execute("select blockid, status from bglblock;")
            results = self.db2.fetchall()
            db2data = {}
            [db2data.update({block.strip():status}) for (block, status) in results]
            #print "db2data:", db2data
            for part in [part for part in candidates if db2data[part.element.get('name')] != 'F']:
                syslog(LOG_ERR, 'Partition %s in inconsistent state' % (part.element.get('name')))
                candidates.remove(part)
            #print "after db2 check"
            #print "candidates: ", [cand.element.get('name') for cand in candidates]
            # deps is recursive dep list / pdeps is single ref 
            deps = {}
            pdeps = {}
            [pdeps.update({part.element.get('name'):part.element.get('deps', "").split(':')}) for part in candidates]
            [pdeps[key].remove('') for key in pdeps.keys() if '' in pdeps[key]]
            for block in [part for part in candidates]:
                deps[block] = pdeps[block.element.get('name')]
                left = pdeps[block.element.get('name')]
                while left:
                    current = left.pop()
                    if not pdeps.has_key(current):
                        candidates.remove(block)
                        left = []
                    else:
                        more = [item for item in pdeps[current] if item not in deps[block]]
                        left += more
                        deps[block] += more
            #print "Deps:", pdeps, deps
            #print "After dep check"
            #print "candidates: ", [cand.element.get('name') for cand in candidates]
            # now candidates are only completely free blocks
            potential = {}
            for job in idlejobs:
                potential[job] = [part for part in candidates if part.CanRun(job)]
                if not potential[job]:
                    del potential[job]
            #print "potential placements:", potential
            placements = []
            while potential:
                newjob = potential.keys()[0]
                location = potential[newjob][0]
                location.PlaceJob(newjob)
                newjob.Place(location)
                placements.append((newjob.element.get('jobid'), location.element.get('name')))
                del potential[newjob]
                # now we need to remove location (and all partitions containing it) from potential lists
                for block in [location] + deps[location]:
                    [potential[job].remove(block) for job, places in potential.iteritems() if block in places]
                for job in potential.keys():
                    if not potential[job]:
                        del potential[job]
                #print "potential", potential
            return placements
        else:
            return []
                
class BGSched(Server):
    '''This scheduler implements a fifo policy'''
    subscriptions = [('queue-manager', '**', '**')]
    __implementation__ = 'BGSched'
    __component__ = 'scheduler'
    __dispatch__ = {'events':'HandleEvent',
                    'GetPartition':'partitions.Get', 'AddPartition':'partitions.Add',
                    'DelPartition':'partitions.Del', 'AddReservation':'AddRes',
                    'DelReservation':'DelRes', 'Set':'Partition_Set'}
    __statefields__ = ['partitions']
    __schedcycle__ = 10

    def __setup__(self):
        self.partitions = PartitionSet()
        self.jobs = []
        self.qmconnect = FailureMode("QM Connection")
        self.lastrun = 0

    def __progress__(self):
        since = time() - self.lastrun
        if since > self.__schedcycle__:
            self.RunQueue()
            self.lastrun = time()
        return 0
            
    def RunQueue(self):
        try:
            handle = self.comm.ClientInit('queue-manager')
            self.comm.SendMessage(handle,
                                  "<get-job><job nodes='*' user='*' jobid='*' state='*' walltime='*' queue='*'/></get-job>")
            jobs = self.comm.RecvMessage(handle)
        except:
            self.qmconnect.Fail()
            return 0
        self.qmconnect.Pass()
        xjobs = XML(jobs)
        active = [job.get('jobid') for job in xjobs.findall(".//job")]
        for job in [j for j in self.jobs if j.element.get('jobid') not in active]:
            syslog(LOG_INFO, "Job %s gone from qm" % (job.element.get('jobid')))
            self.jobs.remove(job)
        # known is jobs that are already registered
        known = [job.element.get('jobid') for job in self.jobs]
        [partition.Free() for partition in self.partitions if partition.job not in known + ['none']]
        newjobs = [job for job in xjobs.findall('.//job') if job.get('jobid') not in known]
        self.jobs.extend([Job(job) for job in newjobs])
        placements = self.partitions.Schedule(xjobs.findall(".//job"))
        #print placements
        for (jobid, part) in placements:
            self.comm.SendMessage(handle,
                                  "<run-job nodelist='%s'><job jobid='%s'/></run-job>" %
                                  (part, jobid))
            self.comm.RecvMessage(handle)
        self.comm.ClientClose(handle)
        return 0

    def AddRes(self, xml, portinfo):
        '''Handler for addition of reservations'''
        return self.partitions.Get(xml, portinfo, self.AddRes_cb)

    def AddRes_cb(self, partition, args):
        '''Callback for reservation addition'''
        partition.AddReservation(args)
    
    def DelRes(self, xml, portinfo):
        '''Handler for reservation deletion'''
        return self.partitions.Get(xml, portinfo, self.DelRes_cb)

    def DelRes_cb(self, partition, args):
        '''Callback for reservation deletion'''
        partition.DelReservation(args['name'])

    def Partition_Set_cb(self, partition, args):
        '''Set partition fields callback'''
        partition.element.attrib.update(args)

    def Partition_Set(self, xml, portinfo):
        '''Set Partition Field'''
        return self.partitions.Get(xml, portinfo, self.Partition_Set_cb)

if __name__ == '__main__':
    from getopt import getopt, GetoptError
    try:
        (opts, arguments) = getopt(sys.argv[1:], 'd', ['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\nbgsched.py [-d] [--daemon <pidfile>]" % (msg)
        raise SystemExit, 1
    daemon = [x[1] for x in opts if x[0] == '--daemon']
    debug = len([x for x in opts if x[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0])
    server = BGSched(debug=debug)
    server.ServeForever()
    

